"""Keep reference to running asyncio tasks"""

import asyncio

from typing import Coroutine


__tasks: set[asyncio.Task] = set()  # to avoid garbage collection by holding reference


def create_task(coroutine: Coroutine, name: str) -> asyncio.Task:
    """Creates task and keeps reference until the task is done

    Argument "name" is optional in asyncio.task(), however we require it
    to make our code easier to debug.
    """
    task = asyncio.create_task(coroutine, name=name)
    __tasks.add(task)
    task.add_done_callback(__tasks.discard)
    return task



"""Load configuration from JSON file"""

import json

from typing import Annotated, Optional, Type

from .exceptions import InvalidConfigFile
from .annotations import get_options, set_options


def loader(config_class: Type[Annotated], filename: Optional[str] = None) -> Annotated:
    """Load configuration from JSON file"""
    config = get_options(config_class)

    if filename is None:
        return set_options(config_class(), config)

    with open(filename, 'r', encoding='ascii') as fp:
        conf = json.load(fp)

    if not isinstance(conf, dict):
        # TODO: add unlimited depth support
        raise InvalidConfigFile('JSON configuration structure must be dict[str, dict[str, any]]')

    for module_name, options in config.items():
        module_conf = config.get(module_name, {})
        for option_name, (option_type, default_value) in options.items():
            try:
                value = option_type(module_conf[option_name])
            except KeyError:
                value = default_value
            except Exception as error:
                value = error

            options[option_name] = value

    return set_options(config_class(), config)



"""Service configuration"""

import os
import sys

from .exceptions import InvalidConfigFile


def sys_argv_or_filenames(*filenames: str) -> str:
    """Return usable configuration filename"""
    if len(sys.argv) == 2:
        filename = sys.argv[1]
        if not os.path.isfile(filename):
            raise InvalidConfigFile(f'Provided filename "{filename}" not found')

        return filename

    for filename in filenames:
        if not os.path.isfile(filename):
            continue

        return filename

    raise InvalidConfigFile(f'None of default filenames found: {", ".join(filenames)}')


__all__ = ["InvalidConfigFile", "sys_argv_or_filenames"]
